@@ -6,7 +6,7 @@
 ;; Keywords: languages, processes
 ;; URL: https://github.com/astoff/drepl
 ;; Package-Requires: ((emacs "29.1") (comint-mime "0.3"))
-;; Version: 0.2
+;; Version: 0.3
 
 ;; This program is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -64,6 +64,10 @@ directory name.  The function should accept one argument ASK
 which determines whether to ask or return nil when in doubt."
   :type '(choice string function))
 
+(defcustom drepl-use-savehist-mode nil
+  "Whether to persist REPL input history using `savehist-mode'."
+  :type 'boolean)
+
 (defvar-local drepl--current nil
   "The dREPL associated to the current buffer.
 In dREPL buffers, this is the dREPL object itself.  In all other
@@ -89,7 +93,9 @@ The id of the last request sent.")
 Alist of (ID . CALLBACK) keeping track of requests sent but not
 yet replied to.")
   (pending nil :documentation "\
-List of requests pending to be sent."))
+List of requests pending to be sent.")
+  (history-variable nil :documentation "\
+Name of the REPL history variable."))
 
 (defun drepl--process (repl)
   "The underlying process of dREPL object REPL."
@@ -108,7 +114,8 @@ DOCSTRING is the docstring of the command NAME.
 EXTRA-SLOTS is a list of slots passed to `cl-defstruct' in
 addition to those of `drepl-base'."
   (let* ((display-name (or display-name (symbol-name name)))
-         (conc-name (intern (format "%s--" name))))
+         (conc-name (intern (format "%s--" name)))
+         (hist-var (intern (format "%s--history" name))))
     `(progn
        (defun ,name ()
          ,(or docstring
@@ -117,13 +124,14 @@ addition to those of `drepl-base'."
          (pop-to-buffer (drepl--get-buffer-create ',name t)
                         display-comint-buffer-action))
        (cl-defstruct (,name
-                      (:include drepl-base)
+                      (:include drepl-base (history-variable ',hist-var))
                       (:copier nil)
                       (:constructor nil)
                       (:constructor ,(intern (format "%screate" conc-name)))
                       (:conc-name ,conc-name))
          ,(format "Structure keeping the state of a %s REPL." display-name)
          ,@extra-slots)
+       (defvar ,hist-var nil ,(format "History list for %s REPLs." display-name))
        (put ',name 'drepl--display-name ,display-name))))
 
 (defun drepl--log-message-1 (&rest args)
@@ -133,7 +141,7 @@ ARGS is the entire argument list of `drepl--log-message'."
     (goto-char (point-max))
     (when-let ((w (get-buffer-window)))
       (set-window-point w (point)))
-    (insert (propertize (format-time-string "[%T] ") 'face 'error)
+    (insert (propertize (format-time-string "[%T] ") 'face 'warning)
             (apply #'format args)
             ?\n)))
 
@@ -155,19 +163,12 @@ The message is formed by calling `format' with STRING and ARGS."
       (lambda (s) (json-serialize s :null-object nil))
     (error "Not implemented")))
 
-(defconst drepl--state-transitions
-  '((eval . rawio))
-  "Alist mapping a REPL operation to a new REPL state.
-The default new state is busy.")
-
 (cl-defgeneric drepl--send-request (repl data)
   "Send request data to REPL.
 REPL must be in `ready' state and transitions to `busy' state.
 DATA is a plist containing the request arguments, as well as :op
 and :id entries."
-  (setf (drepl--status repl) (alist-get (plist-get data :op)
-                                        drepl--state-transitions
-                                        'busy nil #'string-equal))
+  (setf (drepl--status repl) 'busy)
   (let* ((proc (drepl--process repl))
          (maxlen (when (process-tty-name proc)
                    (- comint-max-line-length 3))))
@@ -203,12 +204,11 @@ and this function returns the response data directly."
                result))
     (let* ((id (cl-incf (drepl--last-id repl)))
            (data `(:id ,id :op ,(symbol-name op) ,@args)))
-      (push (cons id callback) (if-let ((reqs (drepl--callbacks repl)))
-                                   (cdr reqs)
-                                 (drepl--callbacks repl)))
+      (push (cons id callback) (drepl--callbacks repl))
       (if (eq 'ready (drepl--status repl))
           (drepl--send-request repl data)
-        (push (cons id data) (drepl--pending repl)))
+        ;; Add a pending request to the end of the queue.
+        (setf (drepl--pending repl) (nconc (drepl--pending repl) (list data))))
       id)))
 
 (defun drepl--osc-handler (_cmd text)
@@ -228,24 +228,22 @@ TEXT is a still unparsed message received from the interpreter."
                                nil))
                      (apply-partially #'drepl--handle-notification
                                       repl))))
-    (when-let ((nextreq (and (eq (drepl--status repl) 'ready)
-                             (pop (drepl--pending repl)))))
-      (drepl--send-request repl nextreq))
-    (when callback
-      (funcall callback data))))
+    (when callback (funcall callback data))
+    (when-let ((next (and (eq (drepl--status repl) 'ready)
+                          (pop (drepl--pending repl)))))
+      (drepl--send-request repl next))))
 
 (cl-defgeneric drepl--handle-notification (repl data)
   "Method called when REPL sends a notification.
 DATA is the content of the message."
   (pcase (alist-get 'op data)
-    ("status" (setf (drepl--status repl)
-                    (intern (alist-get 'status data))))
+    ("status"
+     (setf (drepl--status repl) (intern (alist-get 'status data))))
     ("getoptions"
      (setf (drepl--status repl) 'ready)
      (drepl--set-options repl data))
-    ("log" (drepl--log-message "log:%s: %s"
-                               (buffer-name)
-                               (alist-get 'text data)))))
+    ("log"
+     (message "%s: %s" (buffer-name) (alist-get 'text data)))))
 
 ;;; Buffer association
 
@@ -336,17 +334,33 @@ interactively."
   (or (bounds-of-thing-at-point 'symbol)
       (cons (point) (point))))
 
-(defun drepl--completion-cadidates (repl code offset)
-  "Ask REPL for possible completions of CODE with point at OFFSET."
+(defun drepl--completion-cadidates (repl code pos)
+  "Ask REPL for possible completions of CODE with point at POS."
   (let ((response (while-no-input
                     (drepl--communicate repl 'sync 'complete
                                         :code code
-                                        :offset offset))))
+                                        :pos pos))))
     (when (listp response)
-      (mapcar (lambda (c)
-                (let-alist c
-                  (propertize .text 'drepl--annot .annot)))
-              (alist-get 'candidates response)))))
+      (cons
+       (alist-get 'prefix response)
+       (mapcar (lambda (c)
+                 (or (and (stringp c) c)
+                     (let-alist c
+                       (propertize .text 'drepl--annot .annot))))
+               (alist-get 'candidates response))))))
+
+(defun drepl--adjust-candidates (ours theirs cands)
+  "Adjust completion candidate prefixes.
+OURS is the completion prefix computed by Emacs.
+THEIRS is the completion prefix computed by the REPL.
+CANDS is the candidate list computed by the REPL."
+  (let ((d (- (length theirs) (length ours))))
+    (cond ((or (not theirs) (zerop d)) cands)
+          ((cl-plusp d)
+           (mapcar (lambda (s) (substring s d)) cands))
+          ((cl-minusp d)
+           (let ((p (substring ours 0 (abs d))))
+             (mapcar (lambda (s) (concat p s)) cands))))))
 
 (defun drepl--complete ()
   "Function intended for use as a member of `completion-at-point-functions'."
@@ -356,9 +370,13 @@ interactively."
              (code (buffer-substring-no-properties
                     (cdr comint-last-prompt)
                     (point-max)))
-             (offset (- (point) (cdr comint-last-prompt)))
-             (cands (when (>= offset 0)
-                      (drepl--completion-cadidates repl code offset)))
+             (pos (- (point) (cdr comint-last-prompt)))
+             (reply (when (>= pos 0)
+                      (drepl--completion-cadidates repl code pos)))
+             (cands (drepl--adjust-candidates
+                     (buffer-substring-no-properties (car bounds) (cdr bounds))
+                     (car reply)
+                     (cdr reply)))
              (metadata '(metadata
                          (category . drepl)
                          (annotation-function . drepl--capf-annotate)))
@@ -378,18 +396,29 @@ interactively."
   "Like `comint-send-string', but check the REPL status first.
 If it is `rawio', then simply send the raw STRING to process PROC.
 Otherwise, make an eval request."
-  (let ((repl (with-current-buffer
+  (let ((string (substring-no-properties string))
+        (repl (with-current-buffer
                   (if proc (process-buffer proc) (current-buffer))
                 (drepl--get-repl nil t))))
     (if (not (eq (drepl--status repl) 'rawio))
-        (drepl--eval repl string)
+        (progn
+          (when-let ((hist (drepl--history-variable repl)))
+            (add-to-history hist string comint-input-ring-size))
+            (drepl--eval repl string))
       (drepl--log-message "send raw %s" string)
       (comint-simple-send proc string))))
 
 (defun drepl-eval (code)
   "Evaluate CODE string in the current buffer's REPL."
   (interactive (list (read-from-minibuffer "Evaluate: ")))
-  (drepl--eval (drepl--get-repl nil t) code))
+  (let ((repl (drepl--get-repl nil t)))
+    (with-current-buffer (drepl--buffer repl)
+      (when-let ((last (cdr comint-last-prompt)))
+        (setq comint-last-prompt nil)
+        (save-excursion
+          (goto-char last)
+          (insert-before-markers ?\n))))
+    (drepl--eval repl code)))
 
 (defun drepl-eval-region (start end)
   "Evaluate region in the current buffer's REPL.
@@ -405,14 +434,16 @@ is a dREPL buffer."
   (interactive (list (when (or current-prefix-arg
                                (derived-mode-p 'drepl-mode))
                        (read-buffer "Evaluate buffer: "))))
-  (with-current-buffer (or buffer (current-buffer))
-    (drepl-eval-region (point-min) (point-max))))
+  (let ((repl drepl--current))
+    (with-current-buffer (or buffer (current-buffer))
+      (let ((drepl--current repl))
+        (drepl-eval-region (point-min) (point-max))))))
 
 (defun drepl-send-input-maybe (&optional force) ;Change this to `newline', with opposite logic
   "Like `comint-send-input', but first check if input is complete.
 If the input is incomplete or invalid code and FORCE is nil,
 insert start a continuation line instead."
-  (interactive "P")
+  (interactive "P" drepl-mode)
   (unless (derived-mode-p 'drepl-mode)
     (user-error "Can't run this command here"))
   (let-alist (when-let ((repl (unless force (drepl--get-repl 'ready)))
@@ -461,12 +492,12 @@ See that variable's docstring for a description of CALLBACK."
   (when-let ((repl (when (derived-mode-p 'drepl-mode)
                      (drepl--get-repl 'ready)))
              (start (cdr comint-last-prompt))
-             (offset (- (point) start))
-             (code (when (>= offset 0)
+             (pos (- (point) start))
+             (code (when (>= pos 0)
                      (buffer-substring-no-properties start (point-max))))
              (cb (lambda (data)
                    (apply callback (drepl--format-eldoc repl data)))))
-    (drepl--communicate repl cb 'describe :code code :offset offset)))
+    (drepl--communicate repl cb 'describe :code code :pos pos)))
 
 ;;; REPL restart
 
@@ -540,7 +571,14 @@ activated.  It should start and initialize a Comint process."
     (apply #'make-comint-in-buffer
            (buffer-name buffer)
            buffer
-           program nil switches)))
+           program nil switches))
+  (when-let ((hist (drepl--history-variable repl)))
+    (when drepl-use-savehist-mode
+      (defvar savehist-minibuffer-history-variables)
+      (cl-pushnew hist savehist-minibuffer-history-variables))
+    (setq comint-input-ring (make-ring comint-input-ring-size))
+    (dolist (cmd (take comint-input-ring-size (symbol-value hist)))
+      (ring-insert-at-beginning comint-input-ring cmd))))
 
 (cl-defgeneric drepl--set-options (repl data)
   "Implementation of the `setoptions' operation.
